#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import QUB
from qubtree import Node
import CurveFitting
from math import *

# implements CurveFitting module api

class GaussNLFitter(CurveFitting.NLFitter):
    def __init__(self):
        CurveFitting.NLFitter.__init__(self)
        self.initParams( 3,
                         ('Amp', 1.0, 1.0e-20, 1.0e20),
                         ('Mean', 1.0, -1.0e20, 1.0e20),
                         ('Std', 1.0, 1.0e-20, 1.0e20),
                         ('Area', 0.0) )
        
    def initComponent(self, param):
        minX, maxX = self.minmaxX()
        step = (maxX - minX) / (self.ncomponent + 1)
        param['Mean'].data[0] = maxX - step
        param['Std'].data[0] = step / 3.0

        # steal 1/n-1 of amp from others
        amp = 0.0
        nprev = self.ncomponent
        for i in range(nprev):
            paramp = self.params[i]['Amp'].data
            steal = paramp[0] / (nprev + 1)
            paramp[0] = paramp[0] - steal
            amp = amp + steal
        param['Amp'].data[0] = amp

    def getEqn(self):
        return 'Amp * exp( -(x - Mean)^2 / (2 * Std^2) )'

    def minmaxX(self):
        minX = 1.0e10
        maxX = -1.0e10
        for x in self.X:
            if x < minX:
                minX = x
            if x > maxX:
                maxX = x
        return (minX, maxX)
    
    def setParamsFromData(self):
        minX, maxX = self.minmaxX()
        amp = max(self.Y) / self.ncomponent
        step = (maxX - minX) / (self.ncomponent + 1)
        cur  = minX + step
        for param in self.params:
            param['Amp'].data[0] = amp
            
            mean = param['Mean']
            mean.data[0] = cur
            mean.data[1] = minX
            mean.data[2] = maxX
            
            param['Std'].data[0] = step / 3.0
            cur = cur + step

    def func(self, paramvals, x):
        xdev = x - paramvals[1]
        return paramvals[0] * exp(- xdev * xdev / (2 * paramvals[2] * paramvals[2]))

    def deriv(self, paramvals, x):
        xdev = x - paramvals[1]
        variance = paramvals[2] * paramvals[2]
        if variance == 0.0:
            variance = 1.0e-10
        amp = paramvals[0]
        if amp == 0.0:
            amp = 1.0e-10
        yval = amp * exp(- xdev * xdev / (2 * variance))

        rtn =  [ yval / amp,
                 yval * xdev / variance,
                 yval * xdev * xdev / (variance * paramvals[2]),
                 0.0 ]

        return rtn        

    def onFitDone(self):
        areasum = 0.0
        for param in self.params:
            areasum = areasum + param['Amp'].data[0] * param['Std'].data[0]
        for param in self.params:
            param['Area'].data[0] = param['Amp'].data[0] * param['Std'].data[0] / areasum

CurveFitting.fitters['Gaussian'] = GaussNLFitter


class GaussFitter: # is old and unused
    def __init__(self):
        self.state = Node('GaussFit')
        self.paramsNode = self.state['Params']
        param = self.paramsNode['Param']
        param['Amp'].data = 1.0
        param['Mean'].data = 1.0
        param['Std'].data = 1.0
        param['Area'].data = 0.0
        self.params = [param]
        self.data = Node('FitData')
        self.data['X'].data = 1.0
        self.data['Y'].data = 1.0
        self.__n = 1

    def setNumComponents(self, n):
        while n < len(self.params):
            self.paramsNode.remove( self.params.pop() )
        while n > len(self.params):
            newparam = self.params[0].clone()
            self.paramsNode.append( newparam )
            self.params.append( newparam )

            minX, maxX = self.minmaxX()
            step = (maxX - minX) / (len(self.params) + 1)
            newparam['Mean'].data = maxX - step
            newparam['Std'].data = step / 3.0

    def minmaxX(self):
        minX = 1.0e10
        maxX = -1.0e10
        for x in self.data['X'].data:
            if x < minX:
                minX = x
            if x > maxX:
                maxX = x
        return (minX, maxX)
    
    def autoFill(self):
        minX, maxX = self.minmaxX()
        step = (maxX - minX) / (len(self.params) + 1)
        cur  = minX + step
        for param in self.params:
            param['Mean'].data[0] = cur
            param['Std'].data[0] = step / 3.0
            cur = cur + step

    def getNumComponents(self):
        return len(self.params)

    def useData(self, dataset):
        self.data = dataset
        self.__n = min(dataset['X'].data.count, dataset['Y'].data.count)
        self.autoFill()
        
    def calcY(self, comp_ix, x, nextX=0.0):
        compute = self.state['Compute']
        if compute.data.count < 2:
            compute.data = [x, 0.0]
        else:
            compute.data[0] = x

        QUB.FitGauss_Eval(self.state, self.params[comp_ix])
        return compute.data[1]

    def promptOptions(self):
        pass

    def getCurves(self, curves):
        return CurveFitting.RebuildCurves(self, curves)

    def fit(self, maxiter, precision=0.01, iterCallback=lambda x:x, doPrintStats=1):
        iters = self.state['MaxIter']
        iters.data = [maxiter, 0]
        self.state['Precision'].data = precision
        QUB.FitGauss_Fit(self.state, self.data)

        areasum = 0.0
        for param in self.params:
            areasum = areasum + param['Amp'].data[0] * param['Std'].data[0]
        for param in self.params:
            param['Area'].data[0] = param['Amp'].data[0] * param['Std'].data[0] / areasum

        #res = 0.0
        #for i in range(n):
        #    oneres = Y[i] - (a + b * X[i])
        #    res = res + oneres * oneres

        #if doPrintStats:
        #    QUB.Report('residual: %g' % res)

        iterArg = Node()
        iterArg.data = iters.data[1]
        iterCallback(iterArg)

        try:
            err = self.state['Err'].data[0]
            if err == -1:
                QUB.Report('Fit error: singular hessian matrix')
            elif err == -2:
                QUB.Report("Fit error: too high Marquardt's parameter")
            elif err == -3:
                QUB.Report("Fit error: non-convergence")
        except:
            pass

# CurveFitting.fitters['Gaussian'] = GaussFitter


